#ifndef ALPHABET_HPP_
#define ALPHABET_HPP_

#include <boost/serialization/export.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/version.hpp>

#include <set>

//! this is a helper class for sequence
class Alphabet {
friend class Sequence;
public:
  typedef std::string::const_iterator const_iterator;

  //! case-insensitive test to check a character for existence in our alphabet
  bool exists(const char) const;
  
  // note, if you want to define an alphabet for a sequence, you probably want 
  // to update the enumeration in Sequence, and Sequence::get_sequence
  //! The standard DNA alphabet, with unique, and unknown characters
  static const char *reduced_dna_cstr;
  static const Alphabet &reduced_dna_alphabet();
  //! The standard RNA alphabet, with unique, and unknown characters
  static const char *reduced_rna_cstr;
  static const Alphabet &reduced_rna_alphabet();
  //! The standard DNA/RNA alphabet, with unique, and unknown characters
  static const char *reduced_nucleic_cstr;
  static const Alphabet &reduced_nucleic_alphabet();
  //! this is the general IUPAC alphabet for nucleotides
  static const char *nucleic_cstr;
  static const Alphabet &nucleic_alphabet();
  //! the protein alphabet
  static const char *protein_cstr;  
  static const Alphabet &protein_alphabet(); 
    
private:
  //! what are allowable symbols in our alphabet
  std::string alphabet;
  //! internal variable to make exists() faster
  std::set<std::string::value_type> alphabet_set;
  
  //! some necessary string api access
  Alphabet(const char *a);
  //! allow sequence to copy one alphabet to another (needed when unserializing) 
  void assign(const Alphabet& a);
  const_iterator begin() const { return alphabet.begin(); }
  const_iterator end() const { return alphabet.end(); }
  
  // serialization
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive& ar, const unsigned int /*version*/) {
    ar & BOOST_SERIALIZATION_NVP(alphabet);
    alphabet_set.clear();
    alphabet_set.insert(alphabet.begin(), alphabet.end());
  }   
};

#endif /*ALPHABET_*/
