#include "mussa_exceptions.hpp"
#include "alg/alphabet.hpp"
#include "alg/motif_parser.hpp"

#include <boost/spirit/core.hpp>
#include <boost/spirit/actor/push_back_actor.hpp>
#include <boost/spirit/iterator/file_iterator.hpp>
#include <boost/spirit/utility/chset.hpp>
namespace spirit = boost::spirit;

#include <sstream>
#include <stdexcept>

motif_parser::push_channel::push_channel( motif_parser::ParsedMotifs *parsed_ ) :
  parsed(parsed_)
{
}
  
void motif_parser::push_channel::operator()(float f) const
{
  parsed->channels.push_back(f);
}

motif_parser::push_sequence::push_sequence( ParsedMotifs *parsed_) :
  parsed(parsed_)
{
}
  
template<typename Iterator>
void motif_parser::push_sequence::operator()(
  Iterator start,
  Iterator end) const
{
  std::copy(start, end, std::back_inserter(parsed->sequence));   
}

motif_parser::push_name::push_name( ParsedMotifs *parsed_) :
  parsed(parsed_)
{
}

template<typename Iterator>
void motif_parser::push_name::operator()(
  Iterator start,
  Iterator end) const
{
  std::copy(start, end, std::back_inserter(parsed->name));   
}

motif_parser::push_motif::push_motif( ParsedMotifs *parsed_) :
  parsed(parsed_) {}
    
template<typename Iterator>
void motif_parser::push_motif::operator()(
  Iterator start,
  Iterator end) const
{
  float red, green, blue, alpha;
  Sequence seq(parsed->sequence, Sequence::nucleic_alphabet);
  seq.set_fasta_header(parsed->name);
  
  alpha = 1.0;
  switch (parsed->channels.size()) {
    case 4:
      alpha = parsed->channels[3];
      // note fall through.
  case 3:
    red = parsed->channels[0];
    green = parsed->channels[1];
    blue = parsed->channels[2];
    break;        
  default:
    throw std::runtime_error("wrong number of channels");
    break;
  }
  Color c(red, green, blue, alpha);
  parsed->color_mapper->appendInstanceColor("motif", seq.c_str(), c);
  parsed->motifs.insert(seq);
  
  parsed->sequence.clear();
  parsed->name.clear();
  parsed->channels.clear();
}

motif_parser::ParsedMotifs::ParsedMotifs(
  Mussa::motif_set& motifs_, 
  boost::shared_ptr<AnnotationColors> color_mapper_) :
  motifs(motifs_),
  color_mapper(color_mapper_)
{
}

void motif_parser::ParsedMotifs::parse(const std::string &data)
{
  const char *alphabet = Alphabet::nucleic_cstr;

  // parse our string
  spirit::parse_info<std::string::const_iterator> result;
  result = spirit::parse(data.begin(), data.end(),
     *(
       ( 
        (
         (+spirit::chset<>(alphabet))[motif_parser::push_sequence(this)] >> 
         +spirit::space_p
        ) >>
        !(
          (
            // names can either be letter followed by non-space characters
            (spirit::alpha_p >> *spirit::graph_p)[motif_parser::push_name(this)]
            |
            // or a quoted string
            (
             spirit::ch_p('"') >> 
               (+(~spirit::ch_p('"')))[motif_parser::push_name(this)] >>
             spirit::ch_p('"')
            )
          ) >> +spirit::space_p
        ) >>
        spirit::real_p[motif_parser::push_channel(this)] >> +spirit::space_p >>
        spirit::real_p[motif_parser::push_channel(this)] >> +spirit::space_p >>
        spirit::real_p[motif_parser::push_channel(this)] >> +spirit::space_p >>
        !(spirit::real_p[motif_parser::push_channel(this)] >> +spirit::space_p)
       )[push_motif(this)]
     ));
  if (not result.full) {
    std::stringstream msg;
    msg << "Error at character " << result.length; 
    // erase our potentially broken motif list
    motifs.clear();
    throw motif_load_error(msg.str());
  }
}
