#ifndef _MUSSA_SEQUENCE_H_
#define _MUSSA_SEQUENCE_H_
//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                           ---------- sequence.hh -----------
//                        ----------------------------------------

#include <boost/filesystem/path.hpp>
#include <boost/filesystem/fstream.hpp>

#include <boost/serialization/base_object.hpp>
#include <boost/serialization/export.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/version.hpp>
#include <boost/serialization/vector.hpp>

#include <boost/shared_ptr.hpp>

#include <iostream>

#include "alg/alphabet.hpp"

// Sequence data class

//! Attach annotation information to a sequence track
struct annot
{
  annot();
  annot(int begin, int end, std::string type, std::string name);
  ~annot();
  
  int begin;
  int end;
  std::string type;
  std::string name;

  friend bool operator==(const annot& left, const annot& right);
private:
  // boost::serialization support
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive& ar, const unsigned int /*version*/) {
    ar & BOOST_SERIALIZATION_NVP(begin);
    ar & BOOST_SERIALIZATION_NVP(end);
    ar & BOOST_SERIALIZATION_NVP(type);
    ar & BOOST_SERIALIZATION_NVP(name);
  }
};
BOOST_CLASS_EXPORT(annot);


/* The way that motifs are found currently doesn't really 
 * indicate that the match was a reverse compliment
 */
struct motif : public annot
{
  std::string sequence;

  motif() :  annot(), sequence("") {};
  //! this constructor is for when we're adding motifs to our annotations
  motif(int begin, std::string motif);
  ~motif();

  // boost::serialization support
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive& ar, const unsigned int /*version*/) {
    ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(annot);
    ar & BOOST_SERIALIZATION_NVP(sequence);
  }
};
BOOST_CLASS_EXPORT(motif);

//! the only purpose of this class is that the shared_ptr template 
//! functions need the serialization support to be in-class.
class seq_string : public std::string
{
public:
  typedef std::string::iterator iterator;
  typedef std::string::reverse_iterator reverse_iterator;
  typedef std::string::const_iterator const_iterator;
  typedef std::string::const_reverse_iterator const_reverse_iterator;
private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive& ar, const unsigned int /*version*/) {
    //ar & BOOST_SERIALIZATION_BASE_OBJECT_NVP(std::string);
    ar & boost::serialization::make_nvp("bases",
	   boost::serialization::base_object<std::string>(*this)
    );
  }
};

//! sequence track for mussa.
class Sequence 
{
public:
  typedef std::string::value_type value_type;
  typedef std::string::difference_type difference_type;
  typedef std::string::iterator iterator;
  typedef std::string::reverse_iterator reverse_iterator;
  typedef std::string::const_iterator const_iterator;
  typedef std::string::const_reverse_iterator const_reverse_iterator;
  typedef std::string::reference reference;
  typedef std::string::const_reference const_reference;
  typedef std::string::size_type size_type;
  static const size_type npos = std::string::npos;
  enum strand_type { UnknownStrand, PlusStrand, MinusStrand, BothStrand };
  enum alphabet_ref { reduced_dna_alphabet, reduced_rna_alphabet, reduced_nucleic_alphabet, 
                      nucleic_alphabet, protein_alphabet };
                      
  Sequence(alphabet_ref a = reduced_nucleic_alphabet);
  Sequence(const char* seq, alphabet_ref a = reduced_nucleic_alphabet);
  Sequence(const std::string& seq, alphabet_ref a = reduced_nucleic_alphabet);
  Sequence(const Sequence& seq);
  ~Sequence();
  //! assignment to constant sequences
  Sequence &operator=(const Sequence&);

  friend std::ostream& operator<<(std::ostream&, const Sequence&);
  friend bool operator<(const Sequence&, const Sequence&);
  friend bool operator==(const Sequence&, const Sequence&);
  friend bool operator!=(const Sequence&, const Sequence&);
  const_reference operator[](size_type) const;

  //! set sequence to a (sub)string containing nothing but AGCTN
  void set_filtered_sequence(const std::string& seq,
                             alphabet_ref a, 
                             size_type start=0,
                             size_type count=npos,
                             strand_type strand=UnknownStrand);

  //! retrive element at specific position
  const_reference at(size_type n) const;
  //! clear the sequence and its annotations
  void clear();
  //! return c pointer to the sequence data
  const char *c_str() const;
  //! forward iterator
  const_iterator begin() const;
  //! last iterator
  const_iterator end() const;
  //! is our sequence empty?
  bool empty() const;
  //! find first 
  size_type find_first_not_of(const std::string&, size_type index=0); 
  //! how many base pairs are there in our sequence
  size_type size() const;
  //! alias for size (used by string)
  size_type length() const;
  //! reverse iterator
  const_reverse_iterator rbegin() const;
  //! reverse end iterator
  const_reverse_iterator rend() const;
  //! is our sequence empty?
  //! start position relative to "base" sequence
  size_type start() const;
  //! one past the last position relative to "base" sequence
  size_type stop() const;

  //! return a subsequence, copying over any appropriate annotation
  Sequence subseq(int start=0, int count = std::string::npos);
  //! reverse a character
  std::string create_reverse_map() const;
  //! return a reverse compliment (this needs to be improved?)
  Sequence rev_comp() const;

  //! set sequence (filtered)
  void set_sequence(const std::string &, alphabet_ref);
  //! get sequence
  std::string get_sequence() const;
  //! set species name
  void set_species(const std::string &);
  //! get species name
  std::string get_species() const;
  //! set the fasta header
  void set_fasta_header(std::string header);
  //! get the fasta header
  std::string get_fasta_header() const;
  //! get name (will return the first non-empty, of fasta_header, species)
  std::string get_name() const;
  //! return a reference to whichever alphabet we're currently representing
  const Alphabet& get_alphabet() const; 
  //! return a reference to whichever alphabet we're currently representing
  const Alphabet& get_alphabet(alphabet_ref) const; 
  
  //! load sequence from fasta file using the sequences current alphabet
  void load_fasta(const boost::filesystem::path file_path, int seq_num=1,
                  int start_index=0, int end_index=0);
  //! load sequence AGCT from fasta file
  //! \throw mussa_load_error
  //! \throw sequence_empty_error
  //! \throw sequence_empty_file_error
  void load_fasta(const boost::filesystem::path file_path, 
                  alphabet_ref a, 
                  int seq_num=1, 
		              int start_index=0, int end_index=0);
  void load_fasta(std::istream& file, 
                  int seq_num=1, int start_index=0, int end_index=0);
  //! load sequence from stream
  //! \throw mussa_load_error
  //! \throw sequence_empty_error
  //! \throw sequence_empty_file_error
  void load_fasta(std::istream& file, 
                  alphabet_ref a,
                  int seq_num=1, 
		              int start_index=0, int end_index=0);
  //! load sequence annotations
  //! \throws mussa_load_error 
  void load_annot(const boost::filesystem::path file_path, int start_index, int end_index);
  //! load sequence annotations
  //! \throws mussa_load_error 
  void load_annot(std::fstream& data_stream, int start_index, int end_index);
  //! parse annotation file
  /*! \throws annotation_load_error 
   */
  void parse_annot(std::string data, int start_index=0, int end_index=0);
  //! add an annotation to our list of annotations
  void add_annotation(const annot& a);
  const std::list<annot>& annotations() const;
  const std::list<motif>& motifs() const;

  //! add a motif to our list of motifs
  void add_motif(const Sequence& a_motif);
  //! clear our list of found motifs
  void clear_motifs();
  //! search a sequence for a_motif
  //! \throws motif_normalize_error if there's something wrong with a_motif
  std::vector<int> find_motif(const Sequence& a_motif) const;  
  //! annotate the current sequence with other sequences
  void find_sequences(std::list<Sequence>::iterator start, 
		      std::list<Sequence>::iterator end);
  
  void save(boost::filesystem::fstream &save_file);
  void load_museq(boost::filesystem::path load_file_path, int seq_num); 
  
private:
  //! parent sequence
  Sequence *parent;
  //! hold a shared pointer to our sequence string
  boost::shared_ptr<seq_string> seq;
  //! which alphabet we're using
  alphabet_ref alphabet;
  //! start offset into the sequence
  size_type seq_start;
  //! number of basepairs of the shared sequence we represent
  size_type seq_count;
  //! strand orientation
  strand_type strand;
  //! fasta header
  std::string header;
  //! species name
  std::string species;

  //! store our oldstyle annotations
  std::list<annot> annots;
  //! a seperate list for motifs since we're currently not saving them
  std::list<motif> motif_list;

  void motif_scan(const Sequence& a_motif, std::vector<int> * motif_match_starts) const;
  std::string rc_motif(std::string a_motif) const;
  //! look for a string sequence type and and it to an annotation list
  void add_string_annotation(std::string a_seq, std::string name);
  
  // boost::serialization support
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive& ar, const unsigned int /*version*/) {
    ar & BOOST_SERIALIZATION_NVP(parent);
    ar & BOOST_SERIALIZATION_NVP(seq);
    ar & BOOST_SERIALIZATION_NVP(alphabet);
    ar & BOOST_SERIALIZATION_NVP(seq_start);
    ar & BOOST_SERIALIZATION_NVP(seq_count);
    ar & BOOST_SERIALIZATION_NVP(strand);
    ar & BOOST_SERIALIZATION_NVP(header);
    ar & BOOST_SERIALIZATION_NVP(species);
    ar & BOOST_SERIALIZATION_NVP(annots);
    ar & BOOST_SERIALIZATION_NVP(motif_list);
  }
};
BOOST_CLASS_EXPORT(Sequence);

#endif
